import java.util.concurrent.RecursiveAction;
import java.util.Arrays;

public class PMergeSort extends RecursiveAction {
   private Integer[] data;  // values to be sorted
   private int low, high;   // to be sorted: data[low..high)

   @Override
   protected void compute() {
     if (low >= high - 1) return; // size <= 1: sorted already
     int mid = low + (high - low)/2;    // mid point
       // left and right halves
     PMergeSort left = new PMergeSort(data, low, mid);
     PMergeSort right = new PMergeSort(data, mid, high);
     left.fork();   // fork thread working on left
     right.fork();  // fork thread working on right
     left.join();   // wait for sorted left half
     right.join();  // wait for sorted right half
     merge(mid);    // merge halves
   }

   private void merge(int mid) {
     int left = low, right = mid, k = 0;
     Integer[] merged = new Integer[high - low];
     while (left < mid && right < high) {
       if (data[left] <= data[right]) {
         merged[k] = data[left];
         left++;
       } else {
         merged[k] = data[right];
         right++;
       }
       k++;
     }
     while (left < mid) merged[k++] = data[left++];
     while (right < high) merged[k++] = data[right++];
     for (int i = low; i < high; i++)
       data[i] = merged[i - low];
   }

   @Override
   public String toString() {
     return Arrays.toString(data);
   }

   PMergeSort (Integer[] data, int low, int high) {
     this.data = data; this.low = low; this.high = high;
   }

   public static void main(String[] args) {
      Integer[] toSort = new Integer[] {10, -3, 2, 1, 87, 3, 3, 5, 6, -7};
      PMergeSort sorter = new PMergeSort(toSort, 0, toSort.length);
      invokeAll(sorter);
      System.out.println(sorter);
   }
   /* Possible improvements: 
      - add fixed scratch array for merging (avoiding to create new arrays)
      - make class generic
           * <T extends Comparable<? super T>> 
           * x.compareTo(y) < 0  for  x < 0 
           * since it's impossible to create arrays of generic type,
             pass scratch array to constructor */
}

